@@ -1,8 +1,10 @@
 """An asynchronous RabbitMQ client usable for RPC calls"""
+import functools
 import logging
 import secrets
 import sys
 import threading
+import _thread
 import time
 import typing
 
@@ -43,8 +45,14 @@ class Client:
         :type additional_properties: dict[str, str], optional
         """
         # Get a logger for this client
+        self._data_event_handler = None
+        self._allow_messages = threading.Event()
+        self._stop_event = threading.Event()
+        self._data_event_wait_time = data_processing_wait_time
         if additional_properties is None:
-            additional_properties = {}
+            self._additional_properties = {}
+        else:
+            self._additional_properties = additional_properties
         self._logger = logging.getLogger("amqp_rpc_client")
         # = Check if the Data Source Name is a valid data source name =
         self._logger.debug('Validating the following parameter: "amqp_dsn"')
@@ -56,101 +64,44 @@ class Client:
         self._logger.debug("Checking if the parameter is not empty")
         if len(amqp_dsn.strip()) == 0:
             raise ValueError("The amqp_dsn may not be an empty string")
+        self._amqp_dsn = amqp_dsn
         # = Check finished =
         self._logger.debug('All checks for parameter "amqp_dsn" passed')
         if mute_pika:
             self._logger.debug("Muting the underlying pika library completely")
             logging.getLogger("pika").setLevel("CRITICAL")
-        # Parse the amqp_dsn into preliminary parameters
-        self._connection_parameters = pika.URLParameters(amqp_dsn)
-        # Create a new connection name which is added to the client properties later on
-        self.connection_name = "amqp-rpc-client#" + secrets.token_hex(nbytes=8)
-        self._logger.debug("Created connection name for new connection: %s", self.connection_name)
-        # Combine the additional client properties with those set manually here
-        _client_properties = additional_properties | {
-            "connection_name": self.connection_name,
-            "product": "AMQP-RPC Client",
-            "platform": "Python {}".format(sys.version),
-            "information": "Licensed under the 3-Clause BSD License. See the LICENSE file "
-            "supplied with this library",
-            "copyright": "Copyright (c) Jan Eike Suchard",
-        }
-        self._logger.debug("Setting the following client properties: %s", _client_properties)
-        # Set the client properties to the connection parameters
-        self._connection_parameters.client_properties = _client_properties
-        # Create a new blocking connection
-        self._logger.info("Starting the connection to the message broker")
-        self._logger.debug("Creating a new BlockingConnection")
-        self._connection = pika.BlockingConnection(self._connection_parameters)
-        # Open a new channel to the message broker
-        self._logger.debug("Opening a new channel with the BlockingConnection")
-        self._channel = self._connection.channel()
-        # Create a new queue which is exclusive to this client and will be deleted if the client
-        # disconnects. This queue is used for reading responses
-        self._logger.debug("Declaring a new exclusive auto-deleting queue in the opened channel")
-        self._queue = self._channel.queue_declare("", False, False, True, True)
-        # Save the name generated by the broker as response queue name
-        self._response_queue_name = self._queue.method.queue
-        self._logger.info("Connected to the message broker")
-        # Create an event for stopping the broker
-        self._stop_event = threading.Event()
-        # Create an event for allowing messages to be sent after creating the connection
-        self._allow_messages = threading.Event()
-        # Create a thread which will handle the data events sent by the broker
-        self._logger.debug("Setting up the data handling")
-        self._data_event_wait_time = data_processing_wait_time
-        self._data_event_handler = threading.Thread(target=self._handle_data_events, daemon=True)
-        # Start the thread
-        self._logger.debug("Starting the data handling thread")
-        self._data_event_handler.start()
-        self._allow_messages.wait()
-        self._logger.info("Startup process finished. The client may now be used to send messages")
+        self._connect()
 
     def _handle_data_events(self):
         """Handle new data events and cancel the communication with the message broker if the
         event was set"""
-        # Create a new consumer
-        self._logger.debug("Creating new message consumer")
-        self._consumer = self._channel.basic_consume(
-            queue=self._response_queue_name,
-            on_message_callback=self._handle_new_message,
-            auto_ack=False,
-            exclusive=True,
-        )
         # Process some data events from the start on and allow messages to be sent
-        if self._connection.is_open:
-            self._connection.process_data_events()
-            self._allow_messages.set()
-        # Start polling for new messages indefinitely
+        self._connection.process_data_events(time_limit=0)
+        self._allow_messages.set()
         while not self._stop_event.is_set():
-            # Acquire the internal lock and process possible new data events on the connection
-            self.__messaging_lock.acquire()
-            self._logger.debug("Acquired the messaging lock for processing data events")
-            if self._connection.is_open:
-                self._connection.process_data_events()
-            else:
-                self._logger.error("The connection to the message broker is closed. Stopping the AMQP client")
+            try:
+                self._connection.process_data_events(time_limit=0)
+                self._connection.sleep(self._data_event_wait_time)
+            except pika.exceptions.ConnectionClosedByClient:
+                self._logger.info("The client has closed the connection to the message broker")
+                self._allow_messages.clear()
                 self._stop_event.set()
-            self.__messaging_lock.release()
-            # Sleep for 0.01 seconds before rechecking the stop flag
-            self._logger.debug("Released the messaging lock from processing data events")
-            time.sleep(self._data_event_wait_time)
+            except pika.exceptions.ConnectionClosedByBroker as e:
+                self._logger.warning(f"The message broker has closed the connection to the client: {e.reply_text}")
+                self._allow_messages.clear()
+                self._stop_event.set()
+        self._allow_messages.clear()
 
-        # Since the stop flag was set we will now cancel the consuming process
-        self._logger.info("The stopping event was enabled. Cancelling the message consumer")
-        self._channel.basic_cancel(self._consumer)
-        # Now acquire the messaging lock to stop messages from being sent
-        with self.__messaging_lock:
-            # Close the queue used for responses
-            self._logger.debug("Closing the response queue")
-            self._channel.queue_delete(self._response_queue_name)
-            # Close the channel to the message broker
-            self._logger.debug("Closing the channel open to the message broker")
-            self._channel.close()
-            # Close the connection to the message broker
-            self._logger.debug("Closing the connection open to the message broker")
-            self._connection.close()
-        self._logger.info("Closed the connection to the message broker gracefully")
+    def _got_new_message(
+        self,
+        channel: BlockingChannel,
+        method: Basic.Deliver,
+        properties: BasicProperties,
+        content: bytes,
+    ):
+        self._connection.add_callback_threadsafe(
+            functools.partial(self._handle_new_message, channel, method, properties, content)
+        )
 
     def _handle_new_message(
         self,
@@ -183,11 +134,12 @@ class Client:
                 self._logger.critical(
                     "Error in the messaging events. Unable to find event associated with this correlation id"
                 )
-                raise IndexError("Unable to find Event with the correlation id")
+                self._channel.basic_reject(method.delivery_tag, requeue=True)
             else:
                 self.__events.get(properties.correlation_id).set()
+                self._channel.basic_ack(method.delivery_tag)
 
-    def send(self, content: str, exchange: str, routing_key: str = "") -> str:
+    def send(self, content: str, exchange: str, routing_key: str = "") -> typing.Union[str, None]:
         """Send a message to the exchange and get the created message id
 
         :param routing_key: The routing key used to send the message to the right queue [optional for fanout queues]
@@ -195,67 +147,57 @@ class Client:
         :param exchange:  The exchange in which the message shall be published
         :return: The message id created to identify the request
         """
-        if not self._connection.is_open:
-            raise Exception("The AMQP client is not connected to the message broker")
-        # Create the message id
-        self._logger.debug("Creating a new message id")
-        message_id = secrets.token_urlsafe(nbytes=32)
-        self._logger.debug("Created message id: %s", message_id)
-        # Create a new event for the message and insert it into the dict
-        self._logger.debug("Creating a new event for this message")
-        self.__events.update({message_id: threading.Event()})
-        self._logger.debug("Created a new event for this message")
-        # = Send the message to the message broker =
-        # Acquire the messaging lock to allow this message to be sent
-        self.__messaging_lock.acquire()
-        self._logger.debug("Acquired the messaging lock for sending a message")
-        try:
-            self._channel.basic_publish(
-                exchange=exchange,
-                routing_key=routing_key,
-                body=content.encode("utf-8"),
-                properties=pika.BasicProperties(
-                    reply_to=self._response_queue_name,
-                    correlation_id=message_id,
-                    content_encoding="utf-8",
-                    delivery_mode=pika.delivery_mode.DeliveryMode.Persistent,
-                ),
-            )
-        except pika.exceptions.ChannelWrongStateError as e:
-            self.__messaging_lock.release()
-            self._allow_messages.clear()
-            self._stop_event.set()
-            self._logger.warning(
-                "The channel used for sending the message is in the "
-                "wrong state for sending messages. Opening a new channel",
-                exc_info=e,
-            )
-            self._channel = self._connection.channel()
-            self._queue = self._channel.queue_declare("", False, False, True, True)
-            self._response_queue_name = self._queue.method.queue
-            self._data_event_handler.join()
-            self._logger.debug("Setting up the data handling")
-            self._data_event_handler = threading.Thread(target=self._handle_data_events, daemon=True)
-            # Start the thread
-            self._logger.debug("Starting the data handling thread")
-            self._data_event_handler.start()
-            self._allow_messages.wait()
-            self.__messaging_lock.acquire()
-            self._channel.basic_publish(
-                exchange=exchange,
-                routing_key=routing_key,
-                body=content.encode("utf-8"),
-                properties=pika.BasicProperties(
-                    reply_to=self._response_queue_name,
-                    correlation_id=message_id,
-                    content_encoding="utf-8",
-                    delivery_mode=pika.delivery_mode.DeliveryMode.Persistent,
-                ),
-            )
-        self.__messaging_lock.release()
-        self._logger.debug("Published a new message in the specified exchange")
-        self._logger.debug("Released messaging lock from sending a message")
-        return message_id
+        # Acquire the messaging lock for the whole operation
+        with self.__messaging_lock:
+            if not self._connection.is_open:
+                # Since the connection was not opened connect again
+                self._logger.warning(
+                    "WARN_NO_BROKER_CONNECTION - The client was not connected to the message broker. Trying to connect "
+                    "to the message broker"
+                )
+                self._connect()
+            # Since sending messages is allowed start by creating a new message id
+            message_id = secrets.token_urlsafe(nbytes=32)
+            self._logger.debug("Created new message id: %s", message_id)
+            # Create a new event signalizing if the message has been received
+            self.__events.update({message_id: threading.Event()})
+            self._logger.debug("Created new event for the message to be sent")
+            # Now check if the channel used to send the message is still open
+            if not self._channel.is_open:
+                self._logger.warning(
+                    "WARN_NO_OPEN_CHANNEL - The channel to the message broker is not open. Trying to reconnect "
+                    "to the message broker"
+                )
+                self._connect()
+            if not self._allow_messages.wait(timeout=10.0):
+                self._logger.error(
+                    "ERR_NO_MESSAGES_ALLOWED - Unable to send messages since the initial connection flow has not "
+                    "finished "
+                )
+                return None
+            try:
+                self._channel.basic_publish(
+                    exchange=exchange,
+                    routing_key=routing_key,
+                    body=content.encode("utf-8"),
+                    properties=pika.BasicProperties(
+                        reply_to=self._response_queue_name, correlation_id=message_id, content_encoding="utf-8"
+                    ),
+                )
+                self._logger.debug(
+                    "Published the message with the following properties:\nExchange: %s\nRouting "
+                    "Key: %s\nCorrelation ID:%s\nBody:%s\n",
+                    exchange,
+                    routing_key,
+                    message_id,
+                    content,
+                )
+            except Exception as e:
+                self._logger.exception(
+                    "PUBLISH_ERROR - An exception was thrown during the sending of the message", exc_info=e
+                )
+                raise e
+            return message_id
 
     def get_response(self, message_id: str) -> typing.Optional[bytes]:
         """Get a response from the response list
@@ -298,3 +240,66 @@ class Client:
         self._logger.debug("%s - Found Response for the message", message_id)
         response = self.__responses.pop(message_id, None)
         return response
+
+    def _connect(self):
+        # Parse the amqp_dsn into preliminary parameters
+        self._connection_parameters = pika.URLParameters(self._amqp_dsn)
+        # Create a new connection name which is added to the client properties later on
+        self.connection_name = "amqp-rpc-client#" + secrets.token_hex(nbytes=8)
+        self._logger.debug("Created connection name for new connection: %s", self.connection_name)
+        # Combine the additional client properties with those set manually here
+        _client_properties = self._additional_properties | {
+            "connection_name": self.connection_name,
+            "product": "AMQP-RPC Client",
+            "platform": "Python {}".format(sys.version),
+            "information": "Licensed under the 3-Clause BSD License. See the LICENSE file "
+            "supplied with this library",
+            "copyright": "Copyright (c) Jan Eike Suchard",
+        }
+        self._logger.debug("Setting the following client properties: %s", _client_properties)
+        # Set the client properties to the connection parameters
+        self._connection_parameters.client_properties = _client_properties
+        # Create a new blocking connection
+        self._logger.info("Starting the connection to the message broker")
+        self._logger.debug("Creating a new BlockingConnection")
+        self._connection = pika.BlockingConnection(self._connection_parameters)
+        # Open a new channel to the message broker
+        self._logger.debug("Opening a new channel with the BlockingConnection")
+        self._channel = self._connection.channel()
+        # Create a new queue which is exclusive to this client and will be deleted if the client
+        # disconnects. This queue is used for reading responses
+        self._logger.debug("Declaring a new exclusive auto-deleting queue in the opened channel")
+        self._queue = self._channel.queue_declare("", False, False, True, True)
+        # Save the name generated by the broker as response queue name
+        self._response_queue_name = self._queue.method.queue
+        self._logger.info("Connected to the message broker")
+        self._consumer = self._channel.basic_consume(
+            queue=self._response_queue_name,
+            on_message_callback=self._handle_new_message,
+            auto_ack=False,
+            exclusive=True,
+        )
+        self._stop_event.clear()
+        if self._data_event_handler is None:
+            self._data_event_handler = threading.Thread(
+                target=self._handle_data_events, daemon=True, name=f"EVENT_HANDLER_{secrets.token_hex(nbytes=8)}"
+            )
+            self._data_event_handler.start()
+        else:
+            self._data_event_handler.join()
+            self._data_event_handler = threading.Thread(
+                target=self._handle_data_events, daemon=True, name=f"EVENT_HANDLER_{secrets.token_hex(nbytes=8)}"
+            )
+            self._data_event_handler.start()
+        self._logger.debug("Starting the data handling thread")
+        self._allow_messages.wait()
+        self._logger.info("Startup process finished. The client may now be used to send messages")
+
+    def stop(self):
+        """
+        Stop the client and close all connections to the message broker
+        """
+        # Set the stop event since the thread will handle the disconnection flow
+        self._stop_event.set()
+        self._data_event_handler.join()
+        self._connection.close()
